@@ -7,12 +7,14 @@ extern crate log;
 
 use clap::{App, Arg, ArgMatches};
 use llhd;
-use llhd::opt::{Pass, PassContext};
+use moore_circt::{self as circt, mlir, prelude::*, sys::*};
+// use llhd::opt::{Pass, PassContext};
 use moore::common::score::NodeRef;
 use moore::errors::*;
 use moore::name::Name;
 use moore::score::{ScoreBoard, ScoreContext};
-use moore::svlog::{hir::Visitor as _, QueryDatabase as _};
+use moore::source::Span;
+use moore::svlog::{ast::AcceptVisitor as _, hir::Visitor as _, QueryDatabase as _};
 use moore::*;
 use std::path::Path;
 
@@ -62,6 +64,8 @@ fn main() {
                     "ports",
                     "consts",
                     "insts",
+                    "func-args",
+                    "call-args",
                 ])
                 .global(true),
         )
@@ -144,7 +148,13 @@ fn main() {
                 .long("format")
                 .help("Output format")
                 .takes_value(true)
-                .possible_values(&["llhd", "mlir"]),
+                .possible_values(&["llhd", "mlir", "mlir-native"]),
+        )
+        .arg(
+            Arg::with_name("debug-info")
+                .short("g")
+                .long("debug-info")
+                .help("Emit location information as part of the output"),
         )
         .arg(
             Arg::with_name("INPUT")
@@ -172,6 +182,8 @@ fn main() {
             "ports" => Verbosity::PORTS,
             "consts" => Verbosity::CONSTS,
             "insts" => Verbosity::INSTS,
+            "func-args" => Verbosity::FUNC_ARGS,
+            "call-args" => Verbosity::CALL_ARGS,
             _ => unreachable!(),
         };
     }
@@ -329,6 +341,26 @@ fn score(sess: &Session, matches: &ArgMatches) {
             svlog: &svlog_sb,
         };
         let lib_id = ctx.add_library(lib, &asts);
+
+        // Emit the function argument canonicalization details if requested.
+        for root in ctx.svlog.roots() {
+            if ctx.sess.has_verbosity(Verbosity::FUNC_ARGS) {
+                root.accept(&mut svlog::func_args::FuncArgsVerbosityVisitor::new(
+                    ctx.svlog,
+                ));
+            }
+        }
+
+        // Emit the call argument mappings if requested.
+        for root in ctx.svlog.roots() {
+            if ctx.sess.has_verbosity(Verbosity::CALL_ARGS) {
+                root.accept(&mut svlog::call_mapping::CallMappingVerbosityVisitor::new(
+                    ctx.svlog,
+                ));
+            }
+        }
+
+        // Elaborate what has been requested by the user.
         if let Some(names) = matches.values_of("elaborate") {
             debug!("lib_id = {:?}", lib_id);
             debug!("{:?}", sb);
@@ -487,41 +519,170 @@ fn elaborate_name(
                 svlog::InstVerbosityVisitor::new(ctx.svlog).visit_node_with_id(m, false);
             }
 
-            let mut cg = svlog::CodeGenerator::new(ctx.svlog);
+            // Create an MLIR context and load the dialects we need.
+            let mlir_cx = mlir::OwnedContext::new();
+            mlir_cx.load_dialect(circt::std::dialect());
+            mlir_cx.load_dialect(circt::hw::dialect());
+            mlir_cx.load_dialect(circt::comb::dialect());
+            mlir_cx.load_dialect(circt::llhd::dialect());
+            mlir_cx.load_dialect(circt::seq::dialect());
+
+            // Attach a custom diagnostic handler to the context such that we
+            // can print MLIR diagnostics through Moore's own diagnostics
+            // engine.
+            unsafe {
+                mlirContextAttachDiagnosticHandler(
+                    mlir_cx.raw(),
+                    Some(moore_mlir_diagnostic_handler),
+                    // SAFETY: The session reference in the context must outlive
+                    // the MLIR context, which is the case since that context
+                    // is dropped at the end of this function.
+                    ctx.sess as *const _ as *mut _,
+                    None,
+                );
+            }
+
+            // Create the top-level MLIR module.
+            let mlir_module = circt::ModuleOp::new(*mlir_cx);
+
+            let mut cg = svlog::CodeGenerator::new(ctx.svlog, mlir_module);
+            for root in ctx.svlog.roots() {
+                cg.emit_globals(root)?;
+            }
             cg.emit_module(m)?;
-            let mut module = cg.finalize();
-            let pass_ctx = PassContext;
-            if ctx.sess.opts.opt_level > 0 {
-                llhd::pass::ConstFolding::run_on_module(&pass_ctx, &mut module);
-                // llhd::pass::VarToPhiPromotion::run_on_module(&pass_ctx, &mut module); // broken in llhd 0.13
-                llhd::pass::DeadCodeElim::run_on_module(&pass_ctx, &mut module);
-                llhd::pass::GlobalCommonSubexprElim::run_on_module(&pass_ctx, &mut module);
-                llhd::pass::InstSimplification::run_on_module(&pass_ctx, &mut module);
-                llhd::pass::DeadCodeElim::run_on_module(&pass_ctx, &mut module);
+            let module = cg.finalize();
+            // let mut module = cg.finalize();
+            // let pass_ctx = PassContext;
+            // if ctx.sess.opts.opt_level > 0 {
+            //     llhd::pass::ConstFolding::run_on_module(&pass_ctx, &mut module);
+            //     llhd::pass::VarToPhiPromotion::run_on_module(&pass_ctx, &mut module); // broken in llhd 0.13
+            //     llhd::pass::DeadCodeElim::run_on_module(&pass_ctx, &mut module);
+            //     llhd::pass::GlobalCommonSubexprElim::run_on_module(&pass_ctx, &mut module);
+            //     llhd::pass::InstSimplification::run_on_module(&pass_ctx, &mut module);
+            //     llhd::pass::DeadCodeElim::run_on_module(&pass_ctx, &mut module);
+            // }
+
+            // Verify the module.
+            if !mlir_module.verify() {
+                ctx.sess.emit(DiagBuilder2::bug(
+                    "verification failed for emitted MLIR module",
+                ));
+                // mlir_module.dump();
+                return Err(());
             }
 
             // Decide what format to use for the output.
-            emit_output(matches, ctx, &module)?;
+            emit_output(matches, ctx, &module, mlir_module)?
         }
     }
     Ok(())
 }
 
+/// A custom handler for MLIR diagnostics, which prints them through Moore's own
+/// diagnostic engine.
+unsafe extern "C" fn moore_mlir_diagnostic_handler(
+    diag: MlirDiagnostic,
+    sess: *mut std::ffi::c_void,
+) -> MlirLogicalResult {
+    let sess: &Session = (sess as *const Session).as_ref().unwrap();
+
+    // Map the severity from MLIR to Moore.
+    #[allow(non_upper_case_globals)]
+    let severity = match mlirDiagnosticGetSeverity(diag) {
+        MlirDiagnosticSeverity_MlirDiagnosticError => Severity::Error,
+        MlirDiagnosticSeverity_MlirDiagnosticWarning => Severity::Warning,
+        MlirDiagnosticSeverity_MlirDiagnosticNote | MlirDiagnosticSeverity_MlirDiagnosticRemark => {
+            Severity::Note
+        }
+        _ => panic!("unsupported MLIR diagnostic severity"),
+    };
+
+    // Create the main diagnostic message and attach location information if
+    // possible.
+    let mut d = DiagBuilder2::new(severity, mlir_diagnostic_to_string(diag));
+    let main_loc = mlirDiagnosticGetLocation(diag);
+    if let Some(main_span) = mlir_location_to_span(main_loc) {
+        d = d.span(main_span);
+    }
+
+    // Add additional notes.
+    let mut notes_todo = vec![diag];
+    while let Some(note) = notes_todo.pop() {
+        if note.ptr != diag.ptr {
+            let msg = mlir_diagnostic_to_string(note);
+            let loc = mlirDiagnosticGetLocation(note);
+            d = d.add_note(msg);
+            if !loc.ptr.is_null() && main_loc.ptr != loc.ptr {
+                if let Some(span) = mlir_location_to_span(loc) {
+                    d = d.span(span);
+                }
+            }
+        }
+        let num_notes = mlirDiagnosticGetNumNotes(note);
+        for i in (0..num_notes).rev() {
+            notes_todo.push(mlirDiagnosticGetNote(note, i));
+        }
+    }
+
+    sess.emit(d);
+    MlirLogicalResult { value: 1 }
+}
+
+/// Convert the message in an MLIR diagnostic to a string.
+unsafe fn mlir_diagnostic_to_string(diag: MlirDiagnostic) -> String {
+    unsafe extern "C" fn stringify_callback(string: MlirStringRef, to: *mut std::ffi::c_void) {
+        let to: &mut String = (to as *mut String).as_mut().unwrap();
+        to.push_str(
+            std::str::from_utf8(std::slice::from_raw_parts(
+                string.data as *const _,
+                string.length as usize,
+            ))
+            .expect("utf8 string"),
+        );
+    }
+    let mut string = String::new();
+    mlirDiagnosticPrint(
+        diag,
+        Some(stringify_callback),
+        &mut string as *mut _ as *mut _,
+    );
+    string
+}
+
+unsafe fn mlir_location_to_span(loc: MlirLocation) -> Option<Span> {
+    if loc.ptr.is_null() {
+        return None;
+    }
+    if mlirLocationIsFileLineCol(loc) {
+        let filename = mlirStringRefToStr(mlirFileLineColLocGetFilename(loc), String::from);
+        let line = mlirFileLineColLocGetLine(loc) as usize;
+        let column = mlirFileLineColLocGetColumn(loc) as usize;
+        let source = source::get_source_manager().open(&filename).unwrap();
+        let loc = source::Location::with_line_and_column(source, line, column);
+        Some(loc.into())
+    } else {
+        None
+    }
+}
+
 #[derive(Debug)]
 enum OutputFormat {
     Llhd,
     Mlir,
+    MlirNative,
 }
 
 fn emit_output(
     matches: &ArgMatches,
     ctx: &ScoreContext,
     module: &llhd::ir::Module,
+    mlir_module: circt::ModuleOp,
 ) -> Result<(), ()> {
     // Check if the user has provided an explicit output format.
     let fmt = match matches.value_of("output-format") {
         Some("llhd") => Some(OutputFormat::Llhd),
         Some("mlir") => Some(OutputFormat::Mlir),
+        Some("mlir-native") => Some(OutputFormat::MlirNative),
         Some(x) => {
             ctx.sess.emit(DiagBuilder2::fatal(format!(
                 "unknown output format: `{}`",
@@ -566,6 +727,7 @@ fn emit_output(
     match fmt {
         OutputFormat::Llhd => llhd::assembly::write_module(output, &module),
         OutputFormat::Mlir => llhd::mlir::write_module(output, &module),
+        OutputFormat::MlirNative => mlir_module.print(output, matches.is_present("debug-info")),
     };
     Ok(())
 }
@@ -685,16 +847,18 @@ impl<'a, 'gcx> TypeVerbosityVisitor<'a, 'gcx> {
         }
 
         // Report the type context.
-        if let Some(ty) = self.0.type_context(id, self.1) {
-            println!(
-                "{}: type_context({}) = {}",
-                line,
-                ext,
-                match ty {
-                    svlog::typeck::TypeContext::Type(ty) => format!("{}", ty),
-                    svlog::typeck::TypeContext::Bool => "<bool>".to_string(),
-                }
-            );
+        if let Some(expr) = self.0.ast_for_id(id).as_all().get_expr() {
+            if let Some(ty) = self.0.type_context(svlog::Ref(expr), self.1) {
+                println!(
+                    "{}: type_context({}) = {}",
+                    line,
+                    ext,
+                    match ty {
+                        svlog::typeck::TypeContext::Type(ty) => format!("{}", ty),
+                        svlog::typeck::TypeContext::Bool => "<bool>".to_string(),
+                    }
+                );
+            }
         }
     }
 }
